@@ -5,18 +5,19 @@ import scipy
 import scipy.sparse.linalg
 import scipy.optimize
 
-from josephson_junction_array import Array
-from static_problem import DefaultCurrentPhaseRelation
-from static_problem import StaticConfiguration, StaticProblem
+from pyjjasim.josephson_circuit import Circuit
+from pyjjasim.static_problem import DefaultCPR
+from pyjjasim.static_problem import StaticConfiguration, StaticProblem
 
-__all__ = ["TimeEvolutionProblem", "TimeEvolutionResult"]
-
-DEF_TEMPERATURE_PROFILE = lambda search_T: lambda t: np.interp(t, [0, 0.25, 0.75, 1], [1, 1.1 * search_T, 0.9 * search_T, 0])
+__all__ = ["TimeEvolutionProblem", "TimeEvolutionResult", "AnnealingProblem"]
 
+"""
+Time Evolution Module
+"""
 
 class TimeEvolutionProblem:
     """
-    Define multiple time evolution problems with varying parameters in a Josephson Junction Array circuit.
+    Define multiple time evolution problems with varying parameters in a Josephson Junction Circuit circuit.
 
      - All problems are computed in one call (generally much faster than computing individual problems).
      - Use self.compute() to obtain a TimeEvolutionResult object containing the resulting time evolution.
@@ -24,7 +25,7 @@ class TimeEvolutionProblem:
     Physical parameters:       symbol  default
      - time_step               dt      0.05
      - time_step_count         Nt      1000
-     - current_phase_relation  cp      DefaultCurrentPhaseRelation()
+     - current_phase_relation  cp      DefaultCPR()
 
     Problem space parameters: symbol  default   shape
      - frustration            f       0.0       array broadcast compatible to (Nf, W, Nt)
@@ -57,14 +58,11 @@ class TimeEvolutionProblem:
      - config_at_minus_2            None     StaticConfiguration    initial condition at timestep=-2
                                              or None                 -> set to self.get_static_problem(t=0,n=z).compute()
 
-    Algorithm parameters:
-     - voltage_scheme     "forward"  "forward", "central" or "backward"
-
     Methods
      - compute() -> TimeEvolutionResult
      - get_static_problem(time_step=0) -> StaticProblem
      - get_problem_count()
-     - get_array()
+     - get_circuit()
      - get_time_step()
      - get_time_step_count()
      - get_current_phase_relation()
@@ -77,19 +75,19 @@ class TimeEvolutionProblem:
      - get_store_theta()
      - get_store_voltage()
      - get_store_current()
-     - get_voltage_scheme()
+
 
     """
 
-    def __init__(self, array: Array, time_step=0.05, time_step_count=1000,
-                 current_phase_relation=DefaultCurrentPhaseRelation(),
+    def __init__(self, circuit: Circuit, time_step=0.05, time_step_count=1000,
+                 current_phase_relation=DefaultCPR(),
                  frustration=0.0, current_sources=0.0,
                  voltage_sources=0.0, temperature=0.0,
                  store_time_steps=None, store_theta=True, store_voltage=True, store_current=True,
                  config_at_minus_1: np.ndarray = None,
-                 config_at_minus_2: np.ndarray = None,
-                 voltage_scheme="forward"):
-        self.array = array
+                 config_at_minus_2: np.ndarray = None):
+
+        self.circuit = circuit
         self.time_step = time_step
         self.time_step_count = time_step_count
         self.current_phase_relation = current_phase_relation
@@ -101,7 +99,7 @@ class TimeEvolutionProblem:
         self.problem_count = max(get_prob_cnt(frustration),
                                  get_prob_cnt(current_sources), get_prob_cnt(voltage_sources),
                                  get_prob_cnt(temperature))
-        Nj, Nf, W, Nt = self.array._Nj(), self.array._Nf(), self.get_problem_count(), self.get_time_step_count()
+        Nj, Nf, W, Nt = self.circuit._Nj(), self.circuit._Nf(), self.get_problem_count(), self.get_time_step_count()
         self.frustration = frustration if hasattr(frustration, "__call__") else \
             self._broadcast(np.array(frustration), (Nf, W, Nt))
         self.current_sources = current_sources if hasattr(current_sources, "__call__") else \
@@ -117,16 +115,12 @@ class TimeEvolutionProblem:
         self.store_voltage = store_voltage
         self.store_current = store_current
 
-        self.voltage_scheme_name = voltage_scheme
-        self.voltage_scheme = self._set_voltage_scheme(self.voltage_scheme_name)
         self.config_at_minus_1 = np.zeros((Nj, W), dtype=np.double) if config_at_minus_1 is None else config_at_minus_1
-        if self.array._has_capacitance():
-            self.config_at_minus_2 = np.zeros((Nj, W), dtype=np.double) if config_at_minus_2 is None else config_at_minus_2
-
+        self.config_at_minus_2 = np.zeros((Nj, W), dtype=np.double) if config_at_minus_2 is None else config_at_minus_2
         self.prepared_theta_s = np.zeros((Nj, W), dtype=np.double)
 
     def get_static_problem(self, vortex_configuration, problem_nr=0, time_step=0) -> StaticProblem:
-        return StaticProblem(self.array, current_sources=self._Is(time_step)[:, problem_nr].copy(),
+        return StaticProblem(self.circuit, current_sources=self._Is(time_step)[:, problem_nr].copy(),
                              frustration=self._f(time_step)[:, problem_nr].copy(),
                              vortex_configuration=vortex_configuration,
                              current_phase_relation=self.current_phase_relation)
@@ -134,8 +128,8 @@ class TimeEvolutionProblem:
     def get_problem_count(self):
         return self.problem_count
 
-    def get_array(self) -> Array:
-        return self.array
+    def get_circuit(self) -> Circuit:
+        return self.circuit
 
     def get_time_step(self):
         return self.time_step
@@ -156,7 +150,8 @@ class TimeEvolutionProblem:
         return self.current_sources
 
     def get_net_sourced_current(self, time_step):
-        return 0.5 * np.sum(np.abs((self.get_array()._M().A @ self._Is(time_step))), axis=0)
+        M = self.get_circuit().get_cut_matrix()
+        return 0.5 * np.sum(np.abs((M @ self._Is(time_step))), axis=0)
 
     def get_voltage_sources(self):
         return self.voltage_sources
@@ -176,9 +171,6 @@ class TimeEvolutionProblem:
     def get_store_current(self):
         return self.store_current
 
-    def get_voltage_scheme(self):
-        return self.voltage_scheme_name
-
     def get_time(self):
         return np.arange(self._Nt(), dtype=np.double) * self._dt()
 
@@ -187,12 +179,12 @@ class TimeEvolutionProblem:
 
     def compute(self) -> TimeEvolutionResult:
         """
-        Compute time evolution on an Josephson Array.
+        Compute time evolution on an Josephson Circuit.
 
         Requires an initial configuration; step0_init. Must be StaticConfiguration or None.
         If None; it is set to self.to_static_problem().compute()
 
-        If the array has capacitance, requires a second initial configuration; step1_init. If
+        If the circuit has capacitance, requires a second initial configuration; step1_init. If
         this is set to None, it will be assigned the value of step0_init.
 
         If there is no inductance and algorithm=0; the initial condition must obey A(th-g) = 0.
@@ -201,74 +193,10 @@ class TimeEvolutionProblem:
         to apply projection every 100 timesteps.
 
         """
-
-        out = TimeEvolutionResult(self)
-
-        array = self.array
-        Nj = array._Nj()
-        dt, scheme = self._dt(), self.voltage_scheme
-        sm, s0, sp = scheme[0], scheme[1], scheme[2]
-
-        store_th, store_I, store_V = self.store_theta, self.store_current, self.store_voltage
-
-        shape = (array._Nj(), self.get_problem_count())
-        A, AT = array._A().A, array._AT().A
-        Rv = 1 / (dt * array._R().diag(force_as_vector=True, vector_length=Nj)[:, None])
-        Cv = array._C().diag(force_as_vector=True, vector_length=Nj)[:, None] / (dt ** 2)
-
-        theta_next = self.config_at_minus_1
-
-        if array._has_inductance():
-            L = self.array._L().matrix(Nj)
-            L_sw_fact = scipy.sparse.linalg.factorized(A @ L @ AT)
-
-        if array._has_capacitance():
-            theta = self.config_at_minus_2
-            Cprev, C0, Cnext = Cv + sm * Rv, -2.0 * Cv + s0 * Rv, Cv + sp * Rv
-            Asq_fact = scipy.sparse.linalg.factorized(A @ scipy.sparse.diags(1.0 / Cnext[:, 0], 0) @ AT)
-            for i in range(self._Nt()):
-                Is, T, theta_s, df = self._Is(i), self._T(i), self._theta_s(i), self._df(i)
-
-                fluctuations = ((2.0 * T * Rv) ** 0.5) * np.random.randn(*shape)
-                theta_prev = theta.copy()
-                theta = theta_next.copy()
-                if array._has_inductance():
-                    y = AT @ L_sw_fact(A @ (theta + theta_s + L @ Is) - df)
-                    theta_next = -(self._cp(theta) + fluctuations - Is + C0 * theta + Cprev * theta_prev + y) / Cnext
-                    if self.store_time_steps[i]:
-                        out._update([theta_next if self.store_theta else None,
-                                     (theta * s0 + theta_next * sp + theta_prev * sm) / dt if store_V else None,
-                                     Is - y if store_I else None])
-                else:
-                    x = (self._cp(theta) + fluctuations - Is + C0 * theta + Cprev * theta_prev) / Cnext
-                    theta_next = -x + (AT @ Asq_fact(A @ (x + theta_s) - df)) / Cnext
-                    if self.store_time_steps[i]:
-                        out._update([theta_next if self.store_theta else None,
-                                     (theta * s0 + theta_next * sp + theta_prev * sm) / dt if store_V else None,
-                                     (x + theta_next) * Cnext + Is if store_I else None])
+        if self.get_circuit()._has_mixed_inductance():
+            return time_evolution_algo_1(self)
         else:
-            Cnext = Rv
-            Asq_fact = scipy.sparse.linalg.factorized(A @ scipy.sparse.diags(1.0 / Cnext[:, 0], 0) @ AT)
-            for i in range(self._Nt()):
-                Is, T, theta_s, df = self._Is(i), self._T(i), self._theta_s(i), self._df(i)
-                fluctuations = ((2.0 * T * Rv) ** 0.5) * np.random.randn(*shape)
-                theta = theta_next.copy()
-                if array._has_inductance():
-                    y = AT @ L_sw_fact(A @ (theta + theta_s + L @ Is) - df)
-                    theta_next = theta - (self._cp(theta) + fluctuations - Is + y) / Cnext
-                    if self.store_time_steps[i]:
-                        out._update([theta_next if store_th else None,
-                                     (theta_next - theta) / dt if store_V else None,
-                                     Is - y if store_I else None])
-                else:
-                    x = (self._cp(theta) + fluctuations - Is - Cnext * theta) / Cnext
-                    theta_next = -x + (AT @ Asq_fact(A @ (x - theta_s) + df)) / Cnext
-                    if self.store_time_steps[i]:
-                        out._update([theta_next if self.store_theta else None,
-                                     (theta_next - theta) / dt if store_V else None,
-                                     (x + theta_next) * Cnext + Is if store_I else None])
-
-        return out
+            return time_evolution_algo_0(self)
 
     def __str__(self):
         return "time evolution problem: " + \
@@ -289,31 +217,21 @@ class TimeEvolutionProblem:
         return self.time_step
 
     def _f(self, time_step) -> np.ndarray:  # (Nf, W), read-only
-        return np.broadcast_to(self.frustration(time_step), (self.array._Nf(), self.get_problem_count())) \
+        return np.broadcast_to(self.frustration(time_step), (self.circuit._Nf(), self.get_problem_count())) \
             if hasattr(self.frustration, "__call__") else self.frustration[:, :, time_step]
 
     def _Is(self, time_step) -> np.ndarray: # (Nj, W), read-only
-        return np.broadcast_to(self.current_sources(time_step), (self.array._Nj(), self.get_problem_count()))\
+        return np.broadcast_to(self.current_sources(time_step), (self.circuit._Nj(), self.get_problem_count()))\
             if hasattr(self.current_sources, "__call__") else self.current_sources[:, :, time_step]
 
     def _Vs(self, time_step) -> np.ndarray: # (Nj, W), read-only
-        return np.broadcast_to(self.voltage_sources(time_step), (self.array._Nj(), self.get_problem_count()))\
+        return np.broadcast_to(self.voltage_sources(time_step), (self.circuit._Nj(), self.get_problem_count()))\
             if hasattr(self.voltage_sources, "__call__") else self.voltage_sources[:, :, time_step]
 
     def _T(self, time_step) -> np.ndarray:  # (Nj, W), read-only
-        return np.broadcast_to(self.temperature(time_step), (self.array._Nj(), self.get_problem_count()))\
+        return np.broadcast_to(self.temperature(time_step), (self.circuit._Nj(), self.get_problem_count()))\
             if hasattr(self.temperature, "__call__") else self.temperature[:, :, time_step]
 
-    def _df(self, time_step) -> np.ndarray:  # (Nf, W), read-only
-        Nf = self.get_array()._Nf()
-        area = self.array._area().diag(force_as_vector=True, vector_length=Nf)[:, None]
-        if time_step < 0:
-            f = self.config_at_minus_2.problem._f() if time_step == -2 else self.config_at_minus_1.problem._f()
-            return 2 * np.pi * (-area * f)
-        else:
-            f = self._f(time_step)
-        return 2 * np.pi * (- area * f)
-
     def _theta_s(self, time_step) -> np.ndarray:    # (Nj, W), read-only
         # warning: must be called exactly once at every timestep in order, starting below zero.
         if time_step < 0:
@@ -324,15 +242,15 @@ class TimeEvolutionProblem:
 
     def _cp(self, theta) -> np.ndarray:   # (Nj, W)
         # theta -> (Nj, W)
-        Ic = self.get_array()._Ic().diag(force_as_vector=True, vector_length=self.get_array()._Nj())[:, None]
+        Ic = self.get_circuit()._Ic()[:, None]
         return self.current_phase_relation.eval(Ic, theta)
 
     def _dcp(self, theta) -> np.ndarray:  # (Nj, W)
-        Ic = self.get_array()._Ic().diag(force_as_vector=True, vector_length=self.get_array()._Nj())[:, None]
+        Ic = self.get_circuit()._Ic()[:, None]
         return self.current_phase_relation.d_eval(Ic, theta)
 
     def _icp(self, theta) -> np.ndarray:  # (Nj, W)
-        Ic = self.get_array()._Ic().diag(force_as_vector=True, vector_length=self.get_array()._Nj())[:, None]
+        Ic = self.get_circuit()._Ic()[:, None]
         return self.current_phase_relation.i_eval(Ic, theta)
 
     def _broadcast(self, x, shape):
@@ -340,15 +258,6 @@ class TimeEvolutionProblem:
         x = x.reshape(x_shape + (1,) * (len(shape) - len(x_shape)))
         return np.broadcast_to(x, shape)
 
-    def _set_voltage_scheme(self, voltage_scheme):
-        if voltage_scheme in ("forward", None):
-            return [0, -1, 1]
-        if voltage_scheme == "central":
-            return [-0.5, 0, 0.5]
-        if voltage_scheme == "backward":
-            return [-1, 1, 0]
-        return voltage_scheme
-
     def _to_time_point_mask(self, time_points):
         # time_points:  None (represents all points)
         #               or slice
@@ -366,15 +275,118 @@ class TimeEvolutionProblem:
                 raise ValueError("Invalid store_time_steps; must be None, mask, slice or index array")
         return time_points
 
+def time_evolution_algo_0(problem: TimeEvolutionProblem) -> TimeEvolutionResult:
+    out = TimeEvolutionResult(problem)
+
+    circuit = problem.get_circuit()
+    Nj, W = circuit._Nj(), problem.get_problem_count()
+    dt = problem._dt()
+    if circuit._has_mixed_inductance():
+        raise ValueError("Time evolution algorithm 0 does not work with mixed inductance "
+                         "(some loops have no inductance while others do. Use algorithm 1.")
+    store_th, store_I, store_V = problem.store_theta, problem.store_current, problem.store_voltage
+
+    A = circuit.get_cycle_matrix()
+    AT = A.T
+    Rv = 1 / (dt * circuit._R()[:, None])
+    Cv = circuit._C()[:, None] / (dt ** 2)
+    Cprev, C0, Cnext = Cv, -2.0 * Cv - Rv, Cv + Rv
+
+    if circuit._has_inductance():
+        L = problem.circuit._L()
+        L_sw_fact = scipy.sparse.linalg.factorized(A @ L @ AT)
+    Asq_fact = scipy.sparse.linalg.factorized(A @ scipy.sparse.diags(1.0 / Cnext[:, 0], 0) @ AT)
+
+    theta_next = problem.config_at_minus_1
+    theta = problem.config_at_minus_2
+
+    for i in range(problem._Nt()):
+        Is, T, theta_s, f = problem._Is(i), problem._T(i), problem._theta_s(i), problem._f(i)
+
+        rand = np.random.randn(Nj, W) if i % 3 == 0 else rand[np.random.permutation(Nj), :]
+        fluctuations = ((2.0 * T * Rv) ** 0.5) * rand
+
+        theta_prev = theta.copy()
+        theta = theta_next.copy()
+        if circuit._has_inductance():
+            y = AT @ L_sw_fact(A @ (theta + theta_s + L @ Is) + 2 * np.pi * f)
+            theta_next = -(problem._cp(theta) + fluctuations - Is + C0 * theta + Cprev * theta_prev + y) / Cnext
+            I = Is - y if store_I else None
+        else:
+            x = (problem._cp(theta) + fluctuations - Is + C0 * theta + Cprev * theta_prev) / Cnext
+            theta_next = -x + (AT @ Asq_fact(A @ (x - theta_s) - 2 * np.pi * f)) / Cnext
+            I = (x + theta_next) * Cnext + Is if store_I else None
+        V = (theta_next - theta) / dt if store_V else None
+        if problem.store_time_steps[i]:
+            out._update([theta_next if problem.store_theta else None, V, I])
+    return out
+
+def time_evolution_algo_1(problem: TimeEvolutionProblem) -> TimeEvolutionResult:
+    """
+
+    """
+
+    out = TimeEvolutionResult(problem)
+
+    circuit = problem.circuit
+    Nj, Nf, W = circuit._Nj(),  circuit._Nf(), problem.get_problem_count()
+    dt = problem._dt()
+
+    store_th, store_I, store_V = problem.store_theta, problem.store_current, problem.store_voltage
+
+    A = circuit.get_cycle_matrix()
+    M = circuit._Mr()
+    Rv = 1 / (dt * circuit._R()[:, None])
+    Cv = circuit._C()[:, None] / (dt ** 2)
+    Cprev, C0, Cnext = Cv, -2.0 * Cv - Rv, Cv + Rv
+
+    L = problem.circuit._L()
+    L_mask = circuit._get_mixed_inductance_mask()
+    A1 = A[~L_mask, :]
+    A2 = A[L_mask, :]
+
+    Cnext_mat = scipy.sparse.diags(Cnext[:, 0], 0)
+    matrix = scipy.sparse.vstack([M @ Cnext_mat, A1 @ L @ Cnext_mat, A2]).tocsc()
+    m_fact = scipy.sparse.linalg.factorized(matrix)
+
+    theta = problem.config_at_minus_2
+    theta_next = problem.config_at_minus_1
+
+    for i in range(problem._Nt()):
+        Is, T, theta_s, f = problem._Is(i), problem._T(i), problem._theta_s(i), problem._f(i)
+
+        # optimization to only generate new gaussian noise every three timesteps.
+        # On the other timesteps, the last generated noise is shuffled.
+        rand = np.random.randn(Nj, W) if i % 3 == 0 else rand[np.random.permutation(Nj), :]
+        fluctuations = ((2.0 * T * Rv) ** 0.5) * rand
+
+        theta_prev = theta.copy()
+        theta = theta_next.copy()
+        y = problem._cp(theta) + fluctuations + C0 * theta + Cprev * theta_prev
+        F = np.concatenate([M @ (y - Is), A1 @ (theta + theta_s + L @ y) + 2 * np.pi * f[~L_mask],
+                            A2 @ theta_s + 2 * np.pi * f[L_mask]], axis=0)
+        theta_next = -m_fact(F)
+
+        x = Cprev * theta_prev + C0 * theta + Cnext * theta_next + problem._cp(theta) + fluctuations
+        matrix = scipy.sparse.vstack([M, A @ L]).tocsc()
+        print(scipy.linalg.norm(matrix @ x + np.concatenate([-M @ Is, A @ (theta + theta_s) + 2 * np.pi * f], axis=0)))
+
+        if problem.store_time_steps[i]:
+            out._update([theta_next if problem.store_theta else None,
+                         (theta_next - theta) / dt if store_V else None,
+                         y + Cnext * theta_next if store_I else None])
+    return out
+
+
 class TimeEvolutionResult:
     """
-    Represents data of simulated time evolution(s) on a Josephson array.
+    Represents data of simulated time evolution(s) on a Josephson circuit.
 
-    It is defined by a problem, array and any of the quantities theta, current and voltage.
+    It is defined by a problem, circuit and any of the quantities theta, current and voltage.
     These must be of shape (*problem.get_shape(), Nj, Nt_stored). Here Nj is the junction count
     and Nt_stored is the number of time steps that are stored during simulation.
 
-    One can query several properties of the array configurations:
+    One can query several properties of the circuit configurations:
 
      (property)                           (symbol)           (needs)    (shape)
      - phases                             phi                th         (pr_shape, Nn, Nt_stored)
@@ -412,7 +424,7 @@ class TimeEvolutionResult:
 
     def __init__(self, problem: TimeEvolutionProblem):
         self.problem = problem
-        Nj, W, Nt_s = problem.array._Nj(), self.get_problem_count(), problem._Nt_s()
+        Nj, W, Nt_s = problem.circuit._Nj(), self.get_problem_count(), problem._Nt_s()
         self.theta = np.zeros((Nj, W, Nt_s), dtype=np.double) if problem.store_theta else None
         self.voltage = np.zeros((Nj, W, Nt_s), dtype=np.double) if problem.store_theta else None
         self.current = np.zeros((Nj, W, Nt_s), dtype=np.double) if problem.store_theta else None
@@ -456,68 +468,72 @@ class TimeEvolutionResult:
     def get_problem_count(self):
         return self.problem.get_problem_count()
 
-    def get_array(self) -> Array:
-        return self.problem.get_array()
+    def get_circuit(self) -> Circuit:
+        return self.problem.get_circuit()
 
     def select_static_configuration(self, prob_nr, time_step) -> StaticConfiguration:
         if self.theta is None:
             raise ValueError("Theta not stored; cannot select static configuration.")
-        problem = StaticProblem(self.get_array(), current_sources=self.problem._Is(time_step)[:, prob_nr],
+        problem = StaticProblem(self.get_circuit(), current_sources=self.problem._Is(time_step)[:, prob_nr],
                                 frustration=self.problem._f(time_step)[:, prob_nr],
                                 vortex_configuration=self.get_n(time_step)[:, prob_nr],
                                 current_phase_relation=self.problem.current_phase_relation)
         return StaticConfiguration(problem, self.theta[:, prob_nr, time_step])
 
     def get_phi(self, select_time_points=None) -> np.ndarray:
-        M, Mrsq = self.get_array()._Mr().A, self.get_array()._Mrsq().A
+        M = self.get_circuit()._Mr()
+        Mrsq = M @ M.T
         Z = np.zeros((1, self.get_problem_count()), dtype=np.double)
         func = lambda tp: np.concatenate((scipy.sparse.linalg.spsolve(Mrsq, M @ self._th(tp)), Z), axis=0)
-        return self._select(select_time_points, self.get_array()._Nn(), func)
+        return self._select(select_time_points, self.get_circuit()._Nn(), func)
 
     def get_theta(self, select_time_points=None) -> np.ndarray:
-        return self._select(select_time_points, self.get_array()._Nj(), self._th)
+        return self._select(select_time_points, self.get_circuit()._Nj(), self._th)
 
     def get_n(self, select_time_points=None) -> np.ndarray:
-        func = lambda tp:  -self.get_array()._A().A @ np.round(self._th(tp) / (2.0 * np.pi))
-        return self._select(select_time_points, self.get_array()._Nf(), func).astype(int)
+        A = self.get_circuit().get_cycle_matrix()
+        func = lambda tp:  -A @ np.round(self._th(tp) / (2.0 * np.pi))
+        return self._select(select_time_points, self.get_circuit()._Nf(), func).astype(int)
 
     def get_EJ(self, select_time_points=None) -> np.ndarray:
         func = lambda tp: self.problem._icp(self._th(tp))
-        return self._select(select_time_points, self.get_array()._Nj(), func)
+        return self._select(select_time_points, self.get_circuit()._Nj(), func)
 
     def get_I(self, select_time_points=None) -> np.ndarray:
-        return self._select(select_time_points, self.get_array()._Nj(), self._I)
+        return self._select(select_time_points, self.get_circuit()._Nj(), self._I)
 
     def get_Isup(self, select_time_points=None) -> np.ndarray:
         func = lambda tp: self.problem._cp(self._th(tp))
-        return self._select(select_time_points, self.get_array()._Nj(), func)
+        return self._select(select_time_points, self.get_circuit()._Nj(), func)
 
     def get_J(self, select_time_points=None) -> np.ndarray:
-        func = lambda tp: scipy.sparse.linalg.spsolve(self.get_array()._Asq().A,  self.get_array()._A().A @ self._I(tp))
-        return self._select(select_time_points, self.get_array()._Nf(), func)
+        A = self.get_circuit().get_cycle_matrix()
+        func = lambda tp: scipy.sparse.linalg.spsolve(A @ A.T,  A @ self._I(tp))
+        return self._select(select_time_points, self.get_circuit()._Nf(), func)
 
     def get_flux(self, select_time_points=None) -> np.ndarray:
-        Nj, Nf = self.get_array()._Nj(), self.get_array()._Nf()
-        func = lambda tp: self.get_array()._A().A @ (self.get_array()._L().matrix(Nj) @ self._I(tp))
+        Nj, Nf = self.get_circuit()._Nj(), self.get_circuit()._Nf()
+        A = self.get_circuit().get_cycle_matrix()
+        func = lambda tp: A @ (self.get_circuit()._L() @ self._I(tp))
         return self._select(select_time_points, Nf, func)
 
     def get_EM(self, select_time_points=None) -> np.ndarray:
-        Nj = self.get_array()._Nj()
-        func = lambda tp: 0.5 * self.get_array()._L().matrix(Nj) @ (self._I(tp) ** 2)
+        Nj = self.get_circuit()._Nj()
+        func = lambda tp: 0.5 * self.get_circuit()._L() @ (self._I(tp) ** 2)
         return self._select(select_time_points, Nj, func)
 
     def get_V(self, select_time_points=None):
-        return self._select(select_time_points, self.get_array()._Nj(), self._V)
+        return self._select(select_time_points, self.get_circuit()._Nj(), self._V)
 
     def get_U(self, select_time_points=None):
-        M, Mrsq = self.get_array()._Mr().A, self.get_array()._Mrsq().A
+        M = self.get_circuit()._Mr()
+        Mrsq = M @ M.T
         Z = np.zeros((1, self.get_problem_count()), dtype=np.double)
         func = lambda tp: np.concatenate((scipy.sparse.linalg.spsolve(Mrsq, M @ self._V(tp)), Z), axis=0)
-        return self._select(select_time_points, self.get_array()._Nn(), func)
+        return self._select(select_time_points, self.get_circuit()._Nn(), func)
 
     def get_EC(self, select_time_points=None):
-        Nj = self.get_array()._Nj()
-        C = self.get_array()._C().diag(force_as_vector=True, vector_length=Nj)
+        C, Nj = self.get_circuit()._C(), self.get_circuit()._Nj()
         func = lambda tp: 0.5 * C[:, None] * self._V(tp) ** 2
         return self._select(select_time_points, Nj, func)
 
@@ -537,9 +553,9 @@ class TimeEvolutionResult:
              face_quantity_cmap=None, face_quantity_clim=(0, 1), face_quantity_alpha=1,
              face_quantity_logarithmic_colors=False, figsize=None, title="", **kwargs):
 
-        from array_visualize import ArrayPlot
+        from circuit_visualize import CircuitPlot
 
-        return ArrayPlot(self, time_point=time_point, show_vortices=show_vortices, vortex_diameter=vortex_diameter,
+        return CircuitPlot(self, time_point=time_point, show_vortices=show_vortices, vortex_diameter=vortex_diameter,
                          vortex_color=vortex_color, anti_vortex_color=anti_vortex_color,
                          vortex_alpha=vortex_alpha, show_grid=show_grid, grid_width=grid_width,
                          grid_color=grid_color, grid_alpha=grid_alpha, show_colorbar=show_colorbar,
@@ -575,9 +591,9 @@ class TimeEvolutionResult:
                 face_quantity_logarithmic_colors=False, figsize=None,
                 animate_interval=5, title=""):
 
-        from array_visualize import ArrayMovie
+        from pyjjasim.circuit_visualize import CircuitMovie
 
-        return ArrayMovie(self, problem_nr=problem_nr, time_points=time_points, show_vortices=show_vortices,
+        return CircuitMovie(self, problem_nr=problem_nr, time_points=time_points, show_vortices=show_vortices,
                           vortex_diameter=vortex_diameter, vortex_color=vortex_color,
                           anti_vortex_color=anti_vortex_color, vortex_alpha=vortex_alpha,
                           show_grid=show_grid, grid_width=grid_width,
@@ -605,7 +621,7 @@ class TimeEvolutionResult:
                ("I" + self.current.shape.__str__() + ", ") * (self.current is not None) + \
                ("V" + self.voltage.shape.__str__()) * (self.current is not None) + ")" + \
                "\nproblem: " + self.problem.__str__() + \
-               "\narray: " + self.get_array().__str__()
+               "\ncircuit: " + self.get_circuit().__str__()
 
     def _select(self, select_time_points, N, func):
         select_time_points = np.flatnonzero(self.problem._to_time_point_mask(select_time_points))
@@ -615,55 +631,27 @@ class TimeEvolutionResult:
             out[:, :, i] = func(tp)
         return out
 
+
 class AnnealingProblem:
 
-    def __init__(self, array: Array, time_step=0.5, time_step_count=10**4,
-                 frustration=0.0, current_sources=0, search_T=0.0,
-                 problem_count=1, select_lowest_energy=False):
-        self.array = array
-        self.time_step = time_step
-        self.time_step_count = time_step_count
-        self.current_sources = current_sources
-        self.frustration = frustration
-        self.search_T = search_T
-        self.problem_count = problem_count
-        self.select_lowest_energy = select_lowest_energy
+    """
+    Anneals a circuit by gradually lowering the temperature, with the goal for finding a stationairy
+    state with reasonably low energy. The temperature profile is computed automatically based on the
+    measured vortex mobility during the run.
 
-    def compute(self):
-        """
-        Find low energy state by using a time evolution simulation with a slowly decreasing temperature
-        (called annealing). The duration and temperature profile can be controlled.
-        """
-
-        store_time_steps = np.zeros(self.time_step_count, dtype=bool)
-        store_time_steps[-1] = True
-        T = DEF_TEMPERATURE_PROFILE(self.search_T)(np.linspace(0, 1, self.time_step_count))
-        f = self.frustration * np.ones((1, self.problem_count), dtype=np.double)
-        prob = TimeEvolutionProblem(self.array, time_step_count=self.time_step_count, time_step=self.time_step,
-                                    frustration=f, current_sources=self.current_sources, temperature=T[None, None, :],
-                                    store_current=False, store_voltage=False,
-                                    store_time_steps=store_time_steps)
-        out = prob.compute()
-        vortex_configurations = out.get_n()[..., 0]
-        data = [prob.get_static_problem(vortex_configurations[:, p], problem_nr=0, time_step=0).compute()
-                for p in range(self.problem_count)]
-        configurations = [d[0] for d in data]
-        energies = [np.mean(d[0].get_Etot()) for d in data]
-        print(energies)
-        status = [d[1] for d in data]
-        if self.select_lowest_energy:
-            lowest_index = np.argmin(energies)
-            return vortex_configurations[:, lowest_index], energies[lowest_index], status[lowest_index], configurations[lowest_index]
-        return vortex_configurations, energies, status, configurations
-
-
-class AnnealingProblem2:
-
-    def __init__(self, array: Array, time_step=0.5, interval_steps=10,
+     - Does interval_count iterations of interval_steps timeseteps.
+     - The first iteration is done at T=start_T
+     - During each iteration it measures the average vortex mobility. If the target mobility is exceeded,
+       the temperature is divided by T_factor, otherwise it is multiplied by it.
+     - The target vortex mobility v_t at iter i  is v_t(i) = v * ((N - i)/N) ** 1.5, so goes from v to 0.
+     - At the end it does some steps at T=0 to ensure it is settled in a stationairy state.
+     - vortex mobility is defined as abs(n(i+1) - n(i))) / dt averaged over space and time.
+    """
+    def __init__(self, circuit: Circuit, time_step=0.5, interval_steps=10,
                  interval_count=1000, vortex_mobility=0.001,
                  frustration=0.0, current_sources=0, problem_count=1,
                  start_T=1.0, T_factor=1.03):
-        self.array = array
+        self.circuit = circuit
         self.time_step = time_step
         self.interval_steps = interval_steps
         self.interval_count = interval_count
@@ -675,7 +663,7 @@ class AnnealingProblem2:
         self.T_factor = T_factor
 
     def get_vortex_mobility(self, n):
-        Nf = self.array.face_count()
+        Nf = self.circuit.face_count()
         return np.sum(np.sum(np.abs(np.diff(n, axis=2)), axis=2), axis=0) / (Nf * self.time_step * (self.interval_count - 1))
 
     def temperature_adjustment(self, vortex_mobility, iteration):
@@ -687,13 +675,17 @@ class AnnealingProblem2:
 
     def compute(self):
         """
-        Find low energy state by using a time evolution simulation with a slowly decreasing temperature
-        (called annealing). The duration and temperature profile can be controlled.
+        Execute the annealing procedure. Returns:
+        vortex_configurations   (Nf, problem_count)
+        energies                (problem_count,)
+        status
+        configurations          (problem_count,) list of StaticConfiguration objects
+        temperature_profiles    (interval_count, problem_count)
         """
 
-        f = self.frustration * np.ones((1, self.problem_count), dtype=np.double)
-        th = np.zeros((self.array.junction_count(), self.problem_count))
-        prob = TimeEvolutionProblem(self.array, time_step_count=self.interval_steps, time_step=self.time_step,
+        f = np.atleast_1d(self.frustration)[:, None, None]
+        th = np.zeros((self.circuit.junction_count(), self.problem_count))
+        prob = TimeEvolutionProblem(self.circuit, time_step_count=self.interval_steps, time_step=self.time_step,
                                     frustration=f, current_sources=self.current_sources, temperature=self.T,
                                     store_current=False, store_voltage=False)
         T = np.zeros((self.interval_count, self.problem_count))
@@ -717,7 +709,7 @@ class AnnealingProblem2:
         data = [prob.get_static_problem(vortex_configurations[:, p], problem_nr=0, time_step=0).compute()
                 for p in range(self.problem_count)]
         configurations = [d[0] for d in data]
-        energies = [np.mean(d[0].get_Etot()) for d in data]
+        energies = np.array([np.mean(d[0].get_Etot()) for d in data])
         status = np.array([d[1] for d in data])
         return vortex_configurations, energies, status, configurations, T
 
